//   32rbgDIBNX
//
//  2002/ 9/10,2002/10/24 ˁ@P

#include "windef.h"
#include "define.h"
#include "interfaces.h"
#include "classes.h"
#include <stdexcept>

using namespace std;

D2CDIB32::D2CDIB32() {

	m_pbuDIB=NULL;
	m_csStr="32Bit DIB";
	reset(256,256);
	clear(255,255,255);

}

D2CDIB32::D2CDIB32(I32 i32Arg1,I32 i32Arg2) {

	m_pbuDIB=NULL;
	m_csStr="32Bit DIB";
	reset(i32Arg1,i32Arg2);
	clear(255,255,255);

}

D2CDIB32::D2CDIB32(const D2CDIB32& dib) {

	m_pbuDIB=NULL;
	m_csStr="32Bit DIB";

	assign(&dib);

}

D2CDIB32& D2CDIB32::operator=(const D2CDIB32& dib) {

	if (&dib!=this)
		assign(&dib);

	return *this;

}

bool D2CDIB32::assign(const D2IBitmapReader *pbmrArg) {

	if (pbmrArg==NULL || pbmrArg==this)
		return false;

	if (!reset(pbmrArg->getWidth(),pbmrArg->getHeight()))
		return false;

	I32 i32Width=getWidth(),i32Height=getHeight();

	bool bUpset=false;

	if (pbmrArg->getOrigin()==BITMAPORIGIN_LEFTBOTTOM) {

		setHeight(-i32Height);
		bUpset=true;

	}

	for (I32 i=0;i<i32Height;i++)
		for (I32 j=0;j<i32Width;j++)
			if (bUpset)
				setRGB24(j,i,pbmrArg->getRGB24(j,i32Height-i-1));
			else
				setRGB24(j,i,pbmrArg->getRGB24(j,i));

	return true;

}

D2CDIB32::~D2CDIB32() {

	if (m_pbuDIB!=NULL)
		delete m_pbuDIB;

}

UI32 D2CDIB32::getClassID() const { return CID_D2CDIB32; }

UI32 D2CDIB32::getObjectType() const {

	I32 i32ID=VER_D2_0 << 24;
	i32ID+=TYPE_BITMAP << 16;
	i32ID+=ENCODE_DIB32XRGB8;

	return i32ID;

}

UI8 D2CDIB32::getObjectVersion() const { return VER_D2_0; }
UI8 D2CDIB32::getDataType() const { return TYPE_BITMAP; }
UI16 D2CDIB32::getEncoding() const { return ENCODE_DIB32XRGB8; }

void * D2CDIB32::getInterface(UI32 u32Arg) {

	switch(u32Arg) {

	case IID_D2IObject:
		return (D2IObject *)this;

	case IID_D2IBitmapReader:
		return (D2IBitmapReader *)this;

	case IID_D2IBitmapWriter:
		return (D2IBitmapWriter *)this;

	case IID_D2IBitmap:
		return (D2IBitmap *)this;

	case IID_D2IBitmap32ARGB8:
		return (D2IBitmap32ARGB8*)this;

	}

	return NULL;

}

CSTR D2CDIB32::getCSTR() { return m_csStr; }

bool D2CDIB32::serialize(D2IStreamWriter *pswArg,UI32 u32Arg) {

	I32 i,j,i32Length,i32Width=getWidth(),i32Height=getHeight(),i32Size,i32Written;
	D2CBuffer *pbuTmp;

	if (pswArg==NULL)
		return false;

	switch (u32Arg) {

	case ENCODE_NONE:
	case ENCODE_BMP32XRGB8:

		return (pswArg->write(m_pbuDIB)==m_pbuDIB->getSize());

	case ENCODE_BMP:
	case ENCODE_BMP24RGB8:

		if (i32Width % 4==0)
			i32Length=i32Width*3;
		else
			i32Length=i32Width*3+(4-(i32Width*3) % 4);

		i32Size=54+i32Length*i32Height;

		pbuTmp=new D2CBuffer(i32Size);

		if (pbuTmp->getSize()!=i32Size) {

			delete pbuTmp;
			return false;

		}

		pbuTmp->write(0,m_pbuDIB->getBuffer(),54);
		pbuTmp->write(28,24);

		for (i=0;i<i32Height;i++)
			for (j=0;j<i32Width;j++) {

				pbuTmp->write(54+j*3+i*i32Length+2,getR8(j,i));
				pbuTmp->write(54+j*3+i*i32Length+1,getG8(j,i));
				pbuTmp->write(54+j*3+i*i32Length,getB8(j,i));

			}

		i32Written=pswArg->write(pbuTmp);

		delete pbuTmp;

		return i32Written==i32Size;

	default:
		return false;

	}

}

bool D2CDIB32::deserialize(D2IStreamReader *psrArg,UI32 u32Arg) {

	I32 i,j,i32Length,i32Width,i32Height,i32Size,i32Offset,i32Bpp;
	BUF bfData,bfPixels;

	if (psrArg==NULL || (u32Arg!=ENCODE_NONE && u32Arg!=ENCODE_BMP &&
	    u32Arg!=ENCODE_BMP24RGB8 && u32Arg!=ENCODE_BMP32XRGB8))
		return false;

	BUF bfTmp=(BUF)D2Malloc(6);

	if (psrArg->read(bfTmp,6)!=6 || bfTmp[0]!='B' || bfTmp[1]!='M') {

		D2Free(bfTmp);

		return false;

	}

	i32Size=*((I32 *)(bfTmp+2));

	D2Free(bfTmp-6);

	bfData=(BUF)D2Malloc(i32Size-6);

	if (psrArg->read(bfData,i32Size-6)!=i32Size-6) {

		D2Free(bfData);
		return false;

	}

	i32Bpp=bfData[22];
	i32Width=*((UI32 *)(bfData+12));
	i32Height=*((UI32 *)(bfData+16));

	if ((i32Bpp!=24 && i32Bpp!=32) || *((UI32 *)(bfData+24))!=BI_RGB ||
	    i32Width<1 || i32Height==0) {

		D2Free(bfData);

		return false;

	}

	bool bUpset=false;

	if (i32Height<0) {

		i32Height=-i32Height;
		bUpset=true;

	}

	if (i32Bpp==24 && i32Width % 4==0)
		i32Length=i32Width*3;
	else if (i32Bpp==24)
		i32Length=i32Width*3+(4-((i32Width*3) % 4));
	else
		i32Length=i32Width*4;

	i32Offset=*((UI32 *)(bfData+4));

	if (i32Size<i32Offset+i32Length*i32Height) {

		D2Free(bfData);

		return false;


	}

	if (bUpset)
		reset(i32Width,-i32Height);
	else
		reset(i32Width,i32Height);

	bfPixels=bfData+i32Offset-6;

	switch (i32Bpp) {

	case 24:

		for (i=0;i<i32Height;i++)
			for (j=0;j<i32Width;j++) {

				setB8(j,i,bfPixels[j*3+i*i32Length]);
				setG8(j,i,bfPixels[j*3+i*i32Length+1]);
				setR8(j,i,bfPixels[j*3+i*i32Length+2]);

			}

			return true;

	case 32:

		for (i=0;i<i32Height;i++)
			for (j=0;j<i32Width;j++) {

				setB8(j,i,bfPixels[j*4+i*i32Length]);
				setG8(j,i,bfPixels[j*4+i*i32Length+1]);
				setR8(j,i,bfPixels[j*4+i*i32Length+2]);

			}

			return true;

	default:
		return false;

	}

}

bool D2CDIB32::deserialize(D2IBufferReader *pbrArg,UI32 u32Arg) {

	I32 i,j,i32Length,i32Width,i32Height,i32Size,i32Offset,i32Bpp;
	BUF bfData,bfPixels;

	if (pbrArg==NULL || pbrArg->getSize()<54 ||
	    (u32Arg!=ENCODE_NONE && u32Arg!=ENCODE_BMP &&
	    u32Arg!=ENCODE_BMP24RGB8 && u32Arg!=ENCODE_BMP32XRGB8))
		return false;

	bfData=pbrArg->copy();

	if (bfData==NULL || bfData[0]!='B' || bfData[1]!='M') {

		D2Free(bfData);

		return false;

	}

	i32Size=*((I32 *)(bfData+2));

	i32Width=*((UI32 *)(bfData+18));
	i32Height=*((UI32 *)(bfData+22));
	i32Bpp=bfData[28];

	if ((i32Bpp!=24 && i32Bpp!=32) || *((UI32 *)(bfData+30))!=BI_RGB ||
	    i32Width<1 || i32Height==0) {

		D2Free(bfData);

		return false;

	}

	bool bUpset=false;

	if (i32Height<0) {

		i32Height=-i32Height;
		bUpset=true;

	}

	if (i32Bpp==24 && i32Width % 4==0)
		i32Length=i32Width*3;
	else if (i32Bpp==24)
		i32Length=i32Width*3+(4-((i32Width*3) % 4));
	else
		i32Length=i32Width*4;

	i32Offset=*((UI32 *)(bfData+10));

	if (i32Size<i32Offset+i32Length*i32Height) {

		D2Free(bfData);

		return false;


	}

	if (bUpset)
		reset(i32Width,-i32Height);
	else
		reset(i32Width,i32Height);

	bfPixels=bfData+i32Offset;

	switch (i32Bpp) {

	case 24:

		for (i=0;i<i32Height;i++)
			for (j=0;j<i32Width;j++) {

				setB8(j,i,bfPixels[j*3+i*i32Length]);
				setG8(j,i,bfPixels[j*3+i*i32Length+1]);
				setR8(j,i,bfPixels[j*3+i*i32Length+2]);

			}

			return true;

	case 32:

		for (i=0;i<i32Height;i++)
			for (j=0;j<i32Width;j++) {

				setB8(j,i,bfPixels[j*4+i*i32Length]);
				setG8(j,i,bfPixels[j*4+i*i32Length+1]);
				setR8(j,i,bfPixels[j*4+i*i32Length+2]);

			}

			return true;

	default:
		return false;

	}

}

I32 D2CDIB32::getWidth() const { return m_i32Width; }
I32 D2CDIB32::getHeight() const {
	
	if (m_i32Height<0)
		return -m_i32Height;
	else
		return m_i32Height;

}

UI32 D2CDIB32::getOrigin() const {

	if (m_i32Height<0)
		return BITMAPORIGIN_LEFTBOTTOM;
	else
		return BITMAPORIGIN_LEFTTOP;

}

UI32 D2CDIB32::getRGB24(I32 i32Arg1,I32 i32Arg2) const throw(out_of_range) {

	if (i32Arg1<0 || i32Arg1>=getWidth() || i32Arg2<0 || i32Arg2>=getHeight())
		throw out_of_range("ȍWw");

	return m_pu32Pixels[i32Arg1+i32Arg2*getWidth()];

}

UI8 D2CDIB32::getR8(I32 i32Arg1,I32 i32Arg2) const throw(out_of_range) {

	if (i32Arg1<0 || i32Arg1>=getWidth() || i32Arg2<0 || i32Arg2>=getHeight())
		throw out_of_range("ȍWw");

	UI32 u32Res=(m_pu32Pixels[i32Arg1+i32Arg2*getWidth()] & 0x00ff0000) >> 16;

	return (UI8)u32Res;

}

UI8 D2CDIB32::getG8(I32 i32Arg1,I32 i32Arg2) const throw(out_of_range) {

	if (i32Arg1<0 || i32Arg1>=getWidth() || i32Arg2<0 || i32Arg2>=getHeight())
		throw out_of_range("ȍWw");

	UI32 u32Res=(m_pu32Pixels[i32Arg1+i32Arg2*getWidth()] & 0x0000ff00) >> 8;

	return (UI8)u32Res;

}

UI8 D2CDIB32::getB8(I32 i32Arg1,I32 i32Arg2) const throw(out_of_range) {

	if (i32Arg1<0 || i32Arg1>=getWidth() || i32Arg2<0 || i32Arg2>=getHeight())
		throw out_of_range("ȍWw");

	UI32 u32Res=m_pu32Pixels[i32Arg1+i32Arg2*getWidth()] & 0xff;

	return (UI8)u32Res;

}

bool D2CDIB32::setRGB24(I32 i32Arg1,I32 i32Arg2,UI32 u32Arg) {

	if (i32Arg1<0 || i32Arg1>=getWidth() || i32Arg2<0 || i32Arg2>=getHeight())
		return false;

	m_pu32Pixels[i32Arg1+i32Arg2*getWidth()]=u32Arg & 0x00ffffff;

	return true;

}

bool D2CDIB32::setR8(I32 i32Arg1,I32 i32Arg2,UI8 u8Arg) {

	if (i32Arg1<0 || i32Arg1>=getWidth() || i32Arg2<0 || i32Arg2>=getHeight())
		return false;

	UI32 u32Tmp=0x00ff0000 & ((UI32)u8Arg << 16);
	UI32 u32Tmp2=m_pu32Pixels[i32Arg1+i32Arg2*getWidth()] & 0xff00ffff;
	u32Tmp|=u32Tmp2;

	setRGB24(i32Arg1,i32Arg2,u32Tmp);

	return true;

}

bool D2CDIB32::setG8(I32 i32Arg1,I32 i32Arg2,UI8 u8Arg) {

	if (i32Arg1<0 || i32Arg1>=getWidth() || i32Arg2<0 || i32Arg2>=getHeight())
		return false;

	UI32 u32Tmp=0x0000ff00 & ((UI32)u8Arg << 8);
	UI32 u32Tmp2=m_pu32Pixels[i32Arg1+i32Arg2*getWidth()] & 0xffff00ff;
	u32Tmp|=u32Tmp2;

	setRGB24(i32Arg1,i32Arg2,u32Tmp);

	return true;

}

bool D2CDIB32::setB8(I32 i32Arg1,I32 i32Arg2,UI8 u8Arg) {

	if (i32Arg1<0 || i32Arg1>=getWidth() || i32Arg2<0 || i32Arg2>=getHeight())
		return false;

	UI32 u32Tmp=0x000000ff & u8Arg;
	UI32 u32Tmp2=m_pu32Pixels[i32Arg1+i32Arg2*getWidth()] & 0xffffff00;
	u32Tmp|=u32Tmp2;

	setRGB24(i32Arg1,i32Arg2,u32Tmp);

	return true;

}

bool D2CDIB32::setSize(I32 i32Arg1,I32 i32Arg2) {

	m_i32Width=i32Arg1;
	m_i32Height=i32Arg2;

	return true;

}

bool D2CDIB32::reset(I32 i32Arg1,I32 i32Arg2) {

	if (i32Arg1<1 || i32Arg2<0)
		return false;

	I32	i32Size=54+i32Arg1*i32Arg2*4;

	D2CBuffer *pbuTmp=new D2CBuffer(i32Size);

	if (pbuTmp->getSize()!=i32Size) {

		delete pbuTmp;
		return false;

	}

	setSize(i32Arg1,i32Arg2);

	// BITMAPFILEHEADER
	pbuTmp->write(0,'B');
	pbuTmp->write(1,'M');
	pbuTmp->write(2,(BUF)&i32Size,4);
	pbuTmp->fill(6,0,4);
	pbuTmp->write(10,54);

	// BITMAPINFOHEADER
	pbuTmp->write(14,40);
	pbuTmp->fill(15,0,3);
	pbuTmp->write(18,(BUF)&i32Arg1,4);
	pbuTmp->write(22,(BUF)&i32Arg2,4);
	pbuTmp->write(26,1);
	pbuTmp->write(27,0);
	pbuTmp->write(28,32);
	pbuTmp->write(29,0);

	I32 i32Tmp=BI_RGB;
	pbuTmp->write(30,(BUF)&i32Tmp,4);

	if (m_pbuDIB!=NULL)
		delete m_pbuDIB;

	m_pbuDIB=pbuTmp;

	m_pvHeader=(void *)(m_pbuDIB->getBuffer()+14);
	m_pu32Pixels=(UI32 *)(m_pbuDIB->getBuffer()+54);

	return true;

}

bool D2CDIB32::clear(I32 i32Arg1,I32 i32Arg2,I32 i32Arg3) {

	if (i32Arg1<0)
		i32Arg1=0;
	if (i32Arg1>255)
		i32Arg1=255;

	if (i32Arg2<0)
		i32Arg2=0;
	if (i32Arg2>255)
		i32Arg2=255;

	if (i32Arg3<0)
		i32Arg3=0;
	if (i32Arg3>255)
		i32Arg3=255;

	for (int i=0;i<getHeight();i++)
		for (int j=0;j<getWidth();j++) {

			setR8(j,i,(UI8)i32Arg1);
			setG8(j,i,(UI8)i32Arg2);
			setB8(j,i,(UI8)i32Arg3);

		}

	return true;

}

void * D2CDIB32::getHeader() const {

	return m_pvHeader;

}

void * D2CDIB32::getPixels() const {

	return (void *)m_pu32Pixels;

}

UI32 D2CDIB32::getARGB32(I32 i32Arg1,I32 i32Arg2) const throw(out_of_range) {

	return getRGB24(i32Arg1,i32Arg2);

}

bool D2CDIB32::setARGB32(I32 i32Arg1,I32 i32Arg2,UI32 u32Arg) {

	if (i32Arg1<0 || i32Arg1>=getWidth() || i32Arg2<0 || i32Arg2>=getHeight())
		return false;

	return setRGB24(i32Arg1,i32Arg2,u32Arg);

}

bool D2CDIB32::setA8(I32 i32Arg1,I32 i32Arg2,I32 i32Arg3) { return false; }

void * D2CDIB32::getHeaderPointer() const {

	return m_pvHeader;

}

UI32 * D2CDIB32::getPixelsPointer() const {

	return m_pu32Pixels;;

}

bool D2CDIB32::setWidth(I32 i32Arg) {

	m_i32Width=i32Arg;

	return true;

}

bool D2CDIB32::setHeight(I32 i32Arg) {

	m_i32Height=i32Arg;

	return true;

}
